/**
    \mainpage Misionari a kanibalove
    
	\section Zadani
	    Implementujte vlastni reseni problemu zvaneho Misionari a kanibalove. Pro implementaci pouzijte hledani do sirky
		ve stromu, ktery reprezentuje prostor vsech moznych reseni.
	
	\section Popis problmu
	    Tri misionari a tri kanibalove sjou na jedne strane reky a maji k dispozici jednu lodku, ktera uveze jednoho
	    nebo dva lidi. Ukolem je najit zpusob prepravy vsech sesti lidi na druhou stranu reky tak, aby nikdy nezustala
	    skupina misionaru s ciselnou prevahou kanibalu na jednom miste. Ciselna prevaha neni v lodce mozna, takze
	    zbyvaji pouze oba brehy jako moznosti. Pokud ma do lodky nastoupit cestujuci, je jedno, ktery kanibal ci
	    misionar, na kterem brehu se zacina... apod.
	    
	\section Detaily
		-# Stavy: stav se sklada z usporadane sekvence tri cisel reprezentujicich pocet misionaru, pocet kanibalu
		a pocet lodek na kazdem brehu.
		-# Operatory: z kazdeho stavu existuji mozne operatory: do lodky nastoupi 1 nebo 2 misionari,
		1 nebo 2 kanibalove, nebo 1 misionar a 1 kanibal.
		-# Test cile: dosazni stavu (0,0,0) na vychozim brehu
		-# Cena cesty: pocet prejezdu pres reku
		
	\section Implementace
		Popis implementace je uveden v dokumentaci k souboru misionari.cpp. Nase implementace je znacne nefektivni jak casove (je zbytecne nejdriv generovat strom a nasledneho ho prohledavat) tak prostorove, misto stromu by stacilo ukladat frontu stavu (pricemz by bylo zachovano prohledavani do sirky).
**/


/**
	\file misionari.cpp
	
	Implementace resiciho programu pro problem Misionaru a kanibalu.
	
	Zadani popisuje stav jako dvojici trojic cisel: dva brehy, na kazdem brehu 0-3 misionaru a 0-3 kanibalu + prezence lodky.
	Po blizsim zkoumani ale zjistime ze stav je jednoznacne urcen jednim brehem, tudiz celkovy pocet stavu systemu je pouze 4*4*2=32.
	Vytvorime strom vsech moznych reseni problemu (bez opakovani stejnych stavu).
	Nasledne prohledavanim do sirky najdeme a vypiseme vsechny reseni. Reseni jsou vypisovana ve forme seznamu stavu systemu (pro lepsi citelnost jsou vypisovany oba brehy).
**/

#include <iostream>
#include <deque>
#include <vector>
using namespace std;

// forward deklarace Node;
class Node;

/// \brief Pointer na Node objekt
typedef Node* PNode;

typedef unsigned short int usint;
typedef unsigned long int ulint;

/**
	\brief Trida reprezentujici uzel stromu
	
	Kazdy uzel stromu reprezentuje stav reseni.
	Listy stromu reprezentuji konecne stavy reseni naseho problemu, nas budou zajimat pouze koncove stavy (0,0,0).
**/
class Node
{
	private:
		inline bool CheckSezrani(usint misionari, usint kanibalove)
		{
			return (((misionari >= kanibalove) || (misionari == 0)) &&
			((3-misionari) >= (3-kanibalove) || (misionari == 3)));
		}
	public:


		PNode root;
		vector<PNode> children;

		/// \brief Pocet misionaru na pocatecnim brehu
		usint misionari;
		/// \brief Pocet kanibalu na pocatecnim brehu
		usint kanibalove;
		/// \brief Pocet lodek na pocatecnim brehu
		usint lodka;
		
		/// \brief 32bitovy integer reprezentujici uz navstivene stavy
		ulint visited;

		Node() : root(NULL), misionari(3), kanibalove(3), lodka(1), visited((1 << ((3 << 3) | (3 << 1) | 1))) {};
		Node(PNode koren, usint mis, usint kan, usint lod, ulint vis) : root(koren), misionari(mis), kanibalove(kan), lodka(lod), visited(vis) {};
		
		/** \brief Metoda pro generovani novych stavu
		
		    Nove stavy jsou potomky stavu ze ktereho jsou generovany, pricemz se nemuzeme dostat do stavu, ktery uz byl nasim predkem (primym/neprimym).
		**/
		void GenStates();
		
		/** \brief Metoda pro vypis postupu k vysledku
		
			Metoda vypise postup k dosazeni naseho stavu. Pro prehlednost jsou vypisovany oba brehy.
		**/
		void DumpPath();
};


void Node::DumpPath()
{
	deque<PNode> lifo;
	
	PNode current = this;
	lifo.push_back(current);
	
	while (current->root != NULL)
	{
		current = current->root;
		lifo.push_back(current);
	}
	
	while (lifo.size() > 0)
	{
		cout << "[M: " << lifo.back()->misionari << "][K: " << lifo.back()->kanibalove << "]";
		if (lifo.back()->lodka == 1) { cout << "[lodka]      [     ]"; } else { cout << "[     ]      [lodka]"; }
		cout << "[M: " << 3-lifo.back()->misionari << "][K: " << 3-lifo.back()->kanibalove << "]" << endl;

		lifo.pop_back();
	}
}

void Node::GenStates()
{
		if ((misionari == 0) && (kanibalove == 0) && (lodka == 0)) { return; } // reseni

		//OPERATOR 1 : jeden misionar na druhej breh
		//  mame lodku      mame misionare     misionaru bude vic nez kanibalu, nebo nebudou mit co jist
		if ((lodka == 1) && (misionari > 0) && CheckSezrani(misionari-1,kanibalove))
		{
			// misionaru bude o jednoho min, kanibalu stejne, lodka na druhe strane
			ulint stateBit = (1 << (((misionari-1) << 3) | (kanibalove << 1) | 0));
			
			if ((visited & stateBit) == 0)
			{ children.push_back(new Node(this,misionari-1,kanibalove,0,visited | stateBit)); }
		}
		//      na druhem brehu  je tam misionar    misionaru tam bude vic nez kanibalu, nebo nebudou mit co jist
		else if ((lodka == 0) && (misionari < 3) && CheckSezrani(misionari+1,kanibalove))
		{
			// misionaru bude o jednoho vic, kanibalu bude stejne, lodka tady
			ulint stateBit = (1 << (((misionari+1) << 3) | (kanibalove << 1) | 1));
			
			if ((visited & stateBit) == 0)
			{ children.push_back(new Node(this,misionari+1,kanibalove,1,visited | stateBit)); }
		}
		
		//operator 2 - dva misionari na druhej breh
		//  mame lodku      mame misionare     misionaru bude vic nez kanibalu, nebo nebudou mit co jist
		if ((lodka == 1) && (misionari > 1) && CheckSezrani(misionari-2,kanibalove))
		{
            // misionaru bude o dva min, kanibalu stejne, lodka na druhe strane
			ulint stateBit = (1 << (((misionari-2) << 3) | (kanibalove << 1) | 0));
			
			if ((visited & stateBit) == 0)
			{ children.push_back(new Node(this,misionari-2,kanibalove,0,visited | stateBit)); }
		}
		//      na druhem brehu  je tam misionar    misionaru tam bude vic nez kanibalu, nebo nebudou mit co jist
		else if ((lodka == 0) && (misionari < 2) && CheckSezrani(misionari+2,kanibalove))
		{
            // misionaru bude o dva vic, kanibalu bude stejne, lodka tady
			ulint stateBit = (1 << (((misionari+2) << 3) | (kanibalove << 1) | 1));
			
			if ((visited & stateBit) == 0)
			{ children.push_back(new Node(this,misionari+2,kanibalove,1,visited | stateBit)); }
		}

		//operator 3 - jeden kanibal na druhej breh
		//  mame lodku      mame kanibala       misionaru tam bude vic nez kanibalu, nebo nebudou mit co jist
		if ((lodka == 1) && (kanibalove > 0) && CheckSezrani(misionari,kanibalove-1))
		{
            // kanibalu bude o jednoho min, misionaru stejne, lodka na druhe strane
			ulint stateBit = (1 << ((misionari << 3) | ((kanibalove-1) << 1) | 0));
			
			if ((visited & stateBit) == 0)
			{ children.push_back(new Node(this,misionari,kanibalove-1,0,visited | stateBit)); }
		}
		//      na druhem brehu  je tam kanibal      misionaru bude vic nez kanibalu, nebo nebudou mit co jist
		else if ((lodka == 0) && (kanibalove < 3) && CheckSezrani(misionari,kanibalove+1))
		{
            // kanibalu bude o jednoho vic, misionaru bude stejne, lodka tady
			ulint stateBit = (1 << ((misionari << 3) | ((kanibalove+1) << 1) | 1));
			
			if ((visited & stateBit) == 0)
			{ children.push_back(new Node(this,misionari,kanibalove+1,1,visited | stateBit)); }
		}

		//operator 4 - dva kanibalove na druhej breh
		//  mame lodku      mame kanibala       misionaru tam bude vic nez kanibalu, nebo nebudou mit co jist
		if ((lodka == 1) && (kanibalove > 1) && CheckSezrani(misionari,kanibalove-2))
		{
            // kanibalu bude o dva min, misionaru stejne, lodka na druhe strane
			ulint stateBit = (1 << ((misionari << 3) | ((kanibalove-2) << 1) | 0));
			
			if ((visited & stateBit) == 0)
			{ children.push_back(new Node(this,misionari,kanibalove-2,0,visited | stateBit)); }
		}
		//      na druhem brehu  je tam kanibal      misionaru bude vic nez kanibalu, nebo nebudou mit co jist
		else if ((lodka == 0) && (kanibalove < 2) && CheckSezrani(misionari,kanibalove+2))
		{
            // kanibalu bude o dva vic, misionaru bude stejne, lodka tady
			ulint stateBit = (1 << ((misionari << 3) | ((kanibalove+2) << 1) | 1));
			
			if ((visited & stateBit) == 0)
			{ children.push_back(new Node(this,misionari,kanibalove+2,1,visited | stateBit)); }
		}
		
		//operator 5 - misionar a kanibal na druhej breh
		if ((lodka == 1) && (misionari > 0) && (kanibalove > 0) && CheckSezrani(misionari-1,kanibalove-1))
		{
			ulint stateBit = (1 << (((misionari-1) << 3) | ((kanibalove-1) << 1) | 0));
			
			if ((visited & stateBit) == 0)
			{ children.push_back(new Node(this,misionari-1,kanibalove-1,0,visited | stateBit)); }
		}
		else if ((lodka == 0) && (misionari < 3) && (kanibalove < 3) && CheckSezrani(misionari+1,kanibalove+1))
		{
			ulint stateBit = (1 << (((misionari+1) << 3) | ((kanibalove+1) << 1) | 1));
			
			if ((visited & stateBit) == 0)
			{ children.push_back(new Node(this,misionari+1,kanibalove+1,1,visited | stateBit)); }
		}
		
		//cout << "[" << misionari << "] [" << kanibalove << "] [" << lodka << "] [" << children.size() << "]" << endl;
}

int main()
{
	PNode root = new Node();
	deque<PNode> fronta;
	
	fronta.push_back(root);
	
	while (fronta.size() > 0)
	{
		fronta.front()->GenStates();
		ulint length = fronta.front()->children.size();
		
		for (ulint i = 0; i < length; i++)
		{
			fronta.push_back(fronta.front()->children[i]);
		}
		
		fronta.pop_front();
	}
	
	fronta.push_back(root);
	
	while (fronta.size() > 0)
	{
		ulint length = fronta.front()->children.size();

		if (length == 0)
		{
			if ((fronta.front()->misionari == 0) && (fronta.front()->kanibalove == 0) && (fronta.front()->lodka == 0))
			{ fronta.front()->DumpPath(); cout << endl << endl; }
		}
		else
		{
			for (ulint i = 0; i < length; i++)
			{
				fronta.push_back(fronta.front()->children[i]);
			}
		}
		
		fronta.pop_front();
	}
	
	return 1;
}
